-----------------------------------------------------------------------------------------------
-- Client Lua Script for RocketPig
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"

local RPCore = _G["GeminiPackages"]:GetPackage("RPCore-1.1")
local TooltipHelper = _G["GeminiPackages"]:GetPackage("packetdancer-RocketPig-Tooltips")
 
-----------------------------------------------------------------------------------------------
-- RocketPig Module Definition
-----------------------------------------------------------------------------------------------
local RocketPig = {} 
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
-- e.g. local kiExampleVariableMax = 999
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function RocketPig:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here

    return o
end

function RocketPig:Init()
    Apollo.RegisterAddon(self,true)
end
 

-----------------------------------------------------------------------------------------------
-- RocketPig OnLoad
-----------------------------------------------------------------------------------------------
function RocketPig:OnLoad()
    -- Register handlers for events, slash commands and timer, etc.
    -- e.g. Apollo.RegisterEventHandler("KeyDown", "OnKeyDown", self)
    
    -- load our forms
    self.wndMain = Apollo.LoadForm("RocketPigUI.xml", "RocketPigUIForm", nil, self)
    self.wndMain:Show(false)

	Apollo.RegisterSlashCommand("rpset","OnRocketPigCommand",self)
	Apollo.RegisterSlashCommand("rpget","OnRocketPigCommand",self)
	Apollo.RegisterSlashCommand("rpdbg","OnRocketPigCommand",self)
	Apollo.RegisterSlashCommand("rptst","OnRocketPigCommand",self)
	Apollo.RegisterSlashCommand("rpver","OnRocketPigCommand",self)
	Apollo.RegisterSlashCommand("rpclr","OnRocketPigCommand",self)
	Apollo.RegisterSlashCommand("rpcln","OnRocketPigCommand",self)
	Apollo.RegisterSlashCommand("rpstat","OnRocketPigCommand",self)
	
	Apollo.RegisterSlashCommand("rp","OnRocketPig",self)
	
	Apollo.RegisterEventHandler("UnitCreated","OnUnitCreated",self) 
	Apollo.RegisterEventHandler("UnitDestroyed","OnUnitDestroyed",self) 
	
	Apollo.RegisterTimerHandler("RocketPig_TooltipRefresh","OnTooltipRefresh",self) 
	
	Apollo.RegisterEventHandler("RPCore_TraitChanged","OnTraitChanged",self)

	Apollo.CreateTimer("RocketPig_TooltipRefresh",30,true)
end

function RocketPig:OnSave(eLevel)
	if (eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character) then return nil end 

	return { rpCore = RPCore:CacheAsTable() }
end 

function RocketPig:OnRestore(eLevel, tData)
	if (tData.rpCore ~= nil) then 
		if (RPCore == nil) then RPCore = _G["GeminiPackages"]:GetPackage("RPCore-1.0") end 
		RPCore:LoadFromTable(tData.rpCore)
	end
end

-----------------------------------------------------------------------------------------------
-- RocketPig Functions
-----------------------------------------------------------------------------------------------
-- Define general functions here

function RocketPig:OnConfigure()
	self:OnRocketPig()
end 

-----------------------------------------------------------------------------------------------
-- RocketPigForm Functions
-----------------------------------------------------------------------------------------------
-- when the OK button is clicked
function RocketPig:OnOK()
	self.wndMain:Show(false) -- hide the window
	
	local wndFullname = self.wndMain:FindChild("FullnameEditBox")
	local strFullname = wndFullname:GetText()
	local rpState = 0

	local wndCharTitle = self.wndMain:FindChild("CharTitleEditBox")
	local strCharTitle = wndCharTitle:GetText()
		
	local wndBlurb = self.wndMain:FindChild("BlurbEditBox")
	local strBlurb = wndBlurb:GetText()

	for i = 1, 3 do 
		local wndButton = self.wndMain:FindChild("RoleplayToggle" .. i) 
		rpState = RPCore:SetBitFlag(rpState,i,wndButton:IsChecked())
	end 	
	
	RPCore:SetLocalTrait("fullname",strFullname)
	RPCore:SetLocalTrait("rpflag",rpState)
	RPCore:SetLocalTrait("title",strCharTitle) 
	RPCore:SetLocalTrait("shortdesc", strBlurb)
end

-- when the Cancel button is clicked
function RocketPig:OnCancel()
	self.wndMain:Show(false) -- hide the window
end

function RocketPig:OnRocketPigCommand(strCommand, strParams)
	if (strCommand == "rpset") then 
		local space = string.find(strParams," ")
		if (space == nil) then Print("Syntax: rpset <trait> <value>") return end
		
		local trait = string.sub(strParams,1,space - 1) 
		local data = string.sub(strParams,space + 1)
		
		RPCore:SetLocalTrait(trait, data) 
		Print(string.format("rp: %s = %s", trait, data))
	elseif (strCommand == "rpget") then 
		if (string.len(strParams) == 0) then Print("Syntax: rpget [target] <trait>") return end
		local space = string.find(strParams," ")
		
		local target = nil
		local trait = strParams
		if (space ~= nil) then 
			target = string.sub(strParams,1,space - 1) 
			trait = string.sub(strParams,space + 1)
		end
		
		local data, revision = RPCore:FetchTrait(target, trait) 
		if (data ~= nil) then 
			Print(string.format("rp: %s %s =(%d)= %s", target or GameLib.GetPlayerUnit():GetName(),
				trait, revision or 0, tostring(data)))
		else 
			Print(string.format("rp: %s has no cached %s trait", target or GameLib.GetPlayerUnit():GetName(), trait))
		end
	elseif (strCommand == "rpver") then 
		local rpVersion, rpAddons = RPCore:QueryVersion(strParams)
		
		if (rpVersion == nil) then 
			Print(string.format("rp: %s has no cached version information", strParams))
		else 
			Print(string.format("rp: %s is running RPCore %s", strParams, rpVersion))	
		end
	elseif (strCommand == "rpclr") then 
		Print("rp: Clearing cached data.  Thanks for helping!")
		RPCore.tCachedPlayerData = {}
	elseif (strCommand == "rpcln") then 
		Print("rp: Cleaning stale cached data.")
		RPCore:CleanupCache()
	elseif (strCommand == "rpstat") then 
		local nLocal, nCached, nPlayers = RPCore:Stats()
		Print(string.format("rp: %d local traits, %d remote across %d players",
			nLocal, nCached, nPlayers))
	elseif (strCommand == "rpdbg") then 
		Print(string.format("rp: Setting debug level to %d", tonumber(strParams)))
		RPCore:SetDebugLevel(tonumber(strParams))
	end
end

function RocketPig:OnRocketPig()
	local rpFullname = RPCore:GetLocalTrait("fullname") or GameLib.GetPlayerUnit():GetName()
	local rpState = RPCore:GetLocalTrait("rpflag") or 1
	local rpShortBlurb = RPCore:GetLocalTrait("shortdesc") or ""
	local rpTitle = RPCore:GetLocalTrait("title") or ""
	
	local wndFullname = self.wndMain:FindChild("FullnameEditBox")
	wndFullname:SetText(rpFullname)

	local wndCharTitle = self.wndMain:FindChild("CharTitleEditBox")
	wndCharTitle:SetText(rpTitle)

	local wndShortBlurb = self.wndMain:FindChild("BlurbEditBox")
	wndShortBlurb:SetText(rpShortBlurb)
	
	for i = 1, 3 do 
		local wndButton = self.wndMain:FindChild("RoleplayToggle" .. i) 
		wndButton:SetCheck(RPCore:HasBitFlag(rpState,i)) 
	end 	
	
	self.wndMain:Show(true)
end

function RocketPig:OnUnitCreated(unit) 
	if (unit:IsACharacter() and not unit:IsThePlayer()) then TooltipHelper:GenerateAndCacheTooltipForUnit(unit) end 
end 

function RocketPig:OnUnitDestroyed(unit) 
	TooltipHelper:ExpireTooltipForUnit(unit) 
end 

function RocketPig:OnTooltipRefresh()
	TooltipHelper:RebuildTooltips() 
end 

function RocketPig:OnTraitChanged(tData)
	TooltipHelper:RebuildTooltipForUnitNamed(tData.player)
end

-----------------------------------------------------------------------------------------------
-- RocketPig Instance
-----------------------------------------------------------------------------------------------
local RocketPigInst = RocketPig:new()
RocketPigInst:Init()
